/*
** DINA log parser daemon.
**
** (c) Lev Walkin, 2000
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <syslog.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <sysexits.h>

#include <cost.h>
#include <strfunc.h>
#include <lgba.h>

#include "dina.lpd.h"

void daemonize();

#ifdef	USE_DESIRE_ACCOUNTING
#include <desire.h>
#endif

char *acctservice=NULL;

void usage();
void engine();

char *fname=NULL;

FILE *filefd=NULL;
int posfd=-1;

int maintain=0;

int
main(int ac, char **av) {
	int ch;
	int daemonmode = 0;
	int resleft_set = 0;


	if(ac < 2)
		usage();

	while((ch=getopt(ac, av, "a:r:dm")) != -1)
	switch(ch) {
		case 'a':
			if(acctservice)
				errx(1, "Accounting service already specified.");
			acctservice = optarg;
			break;
		case 'd':
			daemonmode = 1;
			break;
		case 'm':
			maintain=1;
			break;
		case 'r':
			if(resleft_set)
				errx(1, "Resource attribute already specified.");
			resleft = optarg;
			resleft_set++;
			break;
		default: 
			usage();
	};

	ac -= optind;
	av += optind;

	if(ac > 1)
		usage();

	fname = *av;

	if(!fname)
		usage();

	/* Initialize engines */

	if((filefd=fopen(fname, "r")) == NULL)
		errx(EX_NOINPUT, "Unable to open specified file: %s.", fname);

	if((posfd=initFILEpos(fname, filefd, pos_ext)) == -1)
		errx(1, "Unable to initialize file position indicator.");

	if(lc_init(NULL))
		errx(EX_CONFIG, "libcost init failed\n");
 
	if(lgba_init(NULL) || lgba_open()) {
		lgba_perror("Error");
		exit(2);
	};

	lgba_setuser("root");

#ifdef LGBA_NODE
	if((les=chnode(LGBA_NODE)))
		fprintf(stderr, "Error in chnode(): %s.\n", lgba_err2string(les));
#endif

#ifdef	USE_DESIRE_ACCOUNTING
	if(acctservice) {
		if(ctb_session(NULL, acctservice, NULL)) {
			warnx("Unable to open DESIRE session '%s'.\n",
				acctservice);
			exit(2);
		};
	};
#endif

	if(daemonmode)
		daemonize();

	openlog(myname, LOG_CONS, LOG_LOCAL6);
	syslog(LOG_INFO, "DINA %s log parser daemon started: %d.",
		engine_name, getpid());

	engine();

	return 0;
};

void
daemonize() {
	int fd;
	warnx("All OK. Daemonizing.");

	switch(fork()) {
	case -1:
		err(1, "Daemon mode failed");
		exit(2);
	case 0: 
		break;
	default:
		_exit(0);
	};

	setsid();

	if ((fd = open("/dev/null", O_RDWR, 0)) != -1) {
		(void)dup2(fd, STDIN_FILENO);
		(void)dup2(fd, STDOUT_FILENO);
		(void)dup2(fd, STDERR_FILENO);
		if (fd > 2)
			(void)close (fd);
	};

};

void
usage() {
	fprintf(stderr, "%s (c) Lev Walkin, %s"
	"\nUsage: %s [-d] [-m] [-a <acct_service>] [-r <res_attr>] <accounting_log_file>"
	"\n\t-d:\tdetach and become daemon"
	"\n\t-m:\tmaintain DINA in-memory line states internally"
	"\n\t-a:\tuse the specified DESIRE accounting service"
	"\n\t-r:\tuse the specified resource attribute (default: \"%s\")"
	"\n", myname, copyright_years, myname, resleft);
	exit(EX_USAGE);
};


