/*
** DINA RADIUS log parser engine.
**
** (c) Lev Walkin, 2000
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <syslog.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>
#include <err.h>

#include <cost.h>
#include <strfunc.h>
#include <lgba.h>

#include "../dina.lpd.common/dina.lpd.h"

char mybuf[16384];

AcctInfo ai;

#define	AV_NAS_ID	"NAS-IP-Address = "
#define	AV_NAS_PORT	"NAS-Port = "
#define	AV_USERNAME	"User-Name = "
#define	AV_ELAPSED	"Acct-Session-Time = "
#define	AV_IBYTES	"Acct-Output-Octets = "
#define	AV_OBYTES	"Acct-Input-Octets = "
#define	AV_STYPE	"Acct-Status-Type = "
#define	AV_SESSION	"Acct-Multi-Session-Id = "

#define	AV_NAS_ID_S	(sizeof(AV_NAS_ID)-1)
#define	AV_NAS_PORT_S	(sizeof(AV_NAS_PORT)-1)
#define	AV_USERNAME_S	(sizeof(AV_USERNAME)-1)
#define	AV_ELAPSED_S	(sizeof(AV_ELAPSED)-1)
#define	AV_IBYTES_S	(sizeof(AV_IBYTES)-1)
#define	AV_OBYTES_S	(sizeof(AV_OBYTES)-1)
#define	AV_STYPE_S	(sizeof(AV_STYPE)-1)
#define	AV_SESSION_S	(sizeof(AV_SESSION)-1)

void
engine() {
	char *s;
	int len;
	int sync = 0;
	struct timespec rqtp;

	rqtp.tv_sec = 0;
	rqtp.tv_nsec = 250000000;

	memset(&ai, 0, sizeof(ai));

	while(1) {
		s=fgets(mybuf, sizeof(mybuf), filefd);
		len = s?strlen(s):0;
		if(!s || !len || s[len-1] != '\n') {
			nanosleep(&rqtp, NULL);
			clearerr(filefd);
			if(s) {
				fseek(filefd, -len, SEEK_CUR);
			} else {
				if(sync == 0 && ai.filled == 0) {
					adjFILEpos(filefd, posfd, sync=1);
				};
			};
			continue;
		};

		if(*s == '\n') {
			if(ai.filled == FI_OK) {
				account(&ai);
				adjFILEpos(filefd, posfd, sync=0);
			};
			ai.filled = 0;
			continue;
		};

		if(*s >= 'A' && *s <= 'Z') {
			memset(&ai, 0, sizeof(ai));
			ai.timestamp = strtotime(s);
			ai.filled = FI_TIMESTAMP;
		};
		if(*s == '\t') {
			s[len-1] = '\0';	/* Clearing newline */
			s++;
			len -= 2;	/* One bytes behind and one bytes after */

			if(!ai.filled) continue;

			if(!strncmp(s, AV_NAS_ID, AV_NAS_ID_S)) {
				if(ai.filled & FI_NAS_ID) {
					syslog(LOG_WARNING, "'%s' specified twice.", AV_NAS_ID);
					ai.filled = 0;
					continue;
				};
				strncpy(ai.nas_id, s+AV_NAS_ID_S, DI_NAS_ID);
				ai.filled |= FI_NAS_ID;
				continue;
			} else if(!strncmp(s, AV_NAS_PORT, AV_NAS_PORT_S)) {
				if(ai.filled & FI_LINE) {
					syslog(LOG_WARNING, "'%s' specified twice.", AV_NAS_PORT);
					ai.filled = 0;
					continue;
				};
				ai.line = atoi(s+AV_NAS_PORT_S);
				if(ai.line < 1) {
					syslog(LOG_WARNING, "Invalid line value: '%s'.",
						s+AV_NAS_PORT_S);
					ai.filled = 0;
					continue;
				}
				ai.filled |= FI_LINE;
				continue;
			} else if(!strncmp(s, AV_USERNAME, AV_USERNAME_S)) {
				if(ai.filled & FI_USERNAME) {
					syslog(LOG_WARNING, "'%s' specified twice.", AV_NAS_PORT);
					ai.filled = 0;
					continue;
				};
				s[len - 1] = '\0';
				strncpy(ai.username, s + AV_USERNAME_S + 1, DI_USERNAME);
				if(!*ai.username) {
					syslog(LOG_WARNING, "Invalid username value: '%s'.",
						s+AV_USERNAME_S+1);
					ai.filled = 0;
					continue;
				};
				ai.filled |= FI_USERNAME;
				continue;
			} else if(!strncmp(s, AV_STYPE, AV_STYPE_S)) {
				if(ai.filled & FI_SS) {
					syslog(LOG_WARNING, "'%s' specified twice.", AV_STYPE);
					ai.filled = 0;
					continue;
				};
				if(!strncmp(s + AV_STYPE_S, "Stop", 4))
					ai.ss=0;
				else
					ai.ss=1;
				ai.filled |= FI_SS;
				continue;
			} else if(!strncmp(s, AV_ELAPSED, AV_ELAPSED_S)) {
				ai.session_time = atoi(s+AV_ELAPSED_S);
				if(ai.session_time < 0) {
					syslog(LOG_WARNING, "Session time negative!");
					ai.session_time = 0;
				};
				continue;
			} else if(!strncmp(s, AV_IBYTES, AV_IBYTES_S)) {
				ai.ibytes = atoi(s + AV_IBYTES_S);
				if(ai.ibytes < 0) {	
					syslog(LOG_WARNING, "Input bytes counter negative!");
					ai.ibytes = 0;
				};
				continue;
			} else if(!strncmp(s, AV_OBYTES, AV_OBYTES_S)) {
				ai.obytes = atoi(s + AV_OBYTES_S);
				if(ai.obytes < 0) {	
					syslog(LOG_WARNING, "Output bytes counter negative!");
					ai.obytes = 0;
				};
				continue;
			} else if(!strncmp(s, AV_SESSION, AV_SESSION_S)) {
				s += AV_SESSION_S;
				if(*s == '"') s++;
				ai.session_id = (unsigned int)strtol(s, NULL, 16);
				continue;
			};

		};

	};

};







