#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <sysexits.h>
#include <time.h>
#include <libdina.h>

void
usage();

void watch_transitions();

int
main(int ac, char **av) {
	int n=0;

	if(dina_ctl_get(DCTL_IPC_CREATED) == -1)
		errx(1, "Could not initialize DINA engine.");

	if(ac < 2) usage();

	if(!strcmp(av[1], "-c")) {
		ac -= 2;
		av += 2;
		if(ac < 3) usage();
		n = nas_line_set_connected(av[0], atoi(av[1]),
			av[2], (ac>3)?atoi(av[3]):0);

	} else if(!strcmp(av[1], "-d")) {
		ac -= 2;
		av += 2;
		if(ac < 2) usage();
		n = nas_line_set_disconnected(av[0], atoi(av[1]),
			(ac>2)?atoi(av[2]):0);
	} else if(!strcmp(av[1], "-w")) {
		watch_transitions();
	} else {
		usage();
	};

	if(n == 0)
		errx(0, "Success.");

	perror("dinactl");

	return 2;
};

void
usage() {
	fprintf(stderr, "dinactl (c) 2000 Lev Walkin <vlm@spelio.net.ru>"
		"\nUsage: dinactl -c <nas_id> <line> <username>	[<session_id>] # connect line"
		"\nUsage: dinactl -d <nas_id> <line> [<session_id>]	# detach line"
		"\nUsage: dinactl -w	# watch line transitions"
	);
	exit(EX_USAGE);
};

extern int _libdina_v_handler;

void
watch_transitions() {
	int status;
	char buf[sizeof(LineEntry)];
	LineEntryEnvelope le;
	int transitions=0;
	int ignorefirst=1;

	if(dina_next_transition(NULL, 0))
		errx(2, "Can't register transitions handler.");

	fprintf(stderr, "Line transitions handler registered: %d\n",
		_libdina_v_handler);

	while(1) {
		fprintf(stderr, "[Waiting for line transition...]\n");
		if(dina_next_transition(&le, 0))
			errx(1, "Error getting next line transition.");

		if(le.Transition - transitions > 1) {
			if(ignorefirst)
				ignorefirst=0;
			else
				fprintf(stderr, "[WARNING: %d previous transitions missed!]\n",
					le.Transition-transitions);
		};
		transitions=le.Transition;

		printf("Transition number:\t%d\n", le.Transition);
		printf("Date:\t%s", ctime(&le.line.TimeStamp));

		if(!*le.line.UserName) { *le.line.UserName = '?'; status = 0; }
		else { status = 1; };
		printf("Status:\t%s\n", status?"Connected":"Disconnected");

		snprintf(buf, DI_NAS_ID, "%s", le.line.NAS);
		printf("Nas:\t[%s]\n", buf);
		printf("Line:\t[%d] (transitions: %u)\n",
			le.line.LineNumber, le.line.Transitions);
		snprintf(buf, DI_USERNAME, "%s", le.line.UserName);
		printf("User:\t[%s]\n", buf);
		printf("\n");
	};

	exit(0);
};

