#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <err.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/sem.h>
#include <sys/msg.h>

#include <libdina.h>
#include "intern.h"

int _libdina_v_handler_pid = 0;
int _libdina_v_handler = 0;

int
dina_next_transition(LineEntryEnvelope *le, int NoWait) {
	int n;
	struct _msgp {
		long mtype;
		LineEntryEnvelope le;
	} msgp;

	/* Initialize DINA */
	if(_libdina_init())
		return -1;

	/* Register transitions handler. */
	if(_libdina_v_handler_pid != getpid()) {
		u_short array[SHSEM_ITEMS];

		if(_libdina_lock())
			return -2;

		if(semctl(_libdina_sem, SHSEM_ITEMS, GETALL, array) == -1) {
			_libdina_unlock();
			return -2; /* Can't register transitions handler. */
		};

		n = 1;
		if(getuid() && getgid())	/* Limit untrusted processes */
			n = SHSEM_ITEMS << 1;

		for(; n < SHSEM_ITEMS; n++)
			if(array[n] == 0)
				break;

		if(n == SHSEM_ITEMS) {
			_libdina_unlock();
			return -2; /* Can't register transitions handler. */
		};

		{
			struct sembuf op[2] = {
				{ n, 0, 0 },
				{ n, 1, SEM_UNDO }
			};
			/* Bind transitions handler. */
			if(semop(_libdina_sem, op, 2) == -1) {
				_libdina_unlock();
				return -2; /* Can't register transitions handler. */
			};
		};

		_libdina_v_handler_pid = getpid();
		_libdina_v_handler = n;

		_libdina_unlock();
		if(!le) return 0;
	};

rcv:

	n=msgrcv(_libdina_msqid, &msgp, sizeof(LineEntryEnvelope),
		_libdina_v_handler, MSG_NOERROR & NoWait?IPC_NOWAIT:0);

	if(n == -1) {
		if(errno == EINTR)
			goto rcv;
		if(errno == EAGAIN)
			return 1;
		return -1;
	};

	if(n != sizeof(LineEntryEnvelope)) {
		warnx("Receiving message: %d != %d", n, sizeof(LineEntryEnvelope));
		goto rcv;
	};

	if(le)
		memcpy(le, &(msgp.le), sizeof(LineEntryEnvelope));

	return 0;
};
