#include <stdlib.h>
#include <stdio.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/msg.h>

#include <string.h>
#include <errno.h>
#include <time.h>

#include <libdina.h>
#include "intern.h"

DINA_MemoryHeader	*_libdina_mheader=NULL;
LineEntry	*_libdina_lines=NULL;
int _libdina_sem=-1;
int _libdina_msqid=-1;
int _libdina_initialized=0;

int _libdina_rdonly=0;

int
dina_init_readonly() {
	if(_libdina_initialized) {
		errno = EPERM;
		return 0;
	};
	_libdina_rdonly=1;
	return _libdina_init();
};

int
_libdina_init() {
	int sg;
	int mempasses = 0;
	size_t memsize=SHMEM_SIZE;

	if(_libdina_initialized)
		return 0;

	if(_libdina_rdonly) {
		_libdina_sem=semget(SHSEM_KEY, SHSEM_ITEMS, 0);
		if(_libdina_sem == -1)
			return -1;

		_libdina_msqid=msgget(SHMSG_KEY, 0);
		if(_libdina_msqid == -1)
			return -1;
	} else {
		_libdina_sem=semget(SHSEM_KEY, SHSEM_ITEMS, 0660 | IPC_CREAT);
		if(_libdina_sem == -1)
			return -1;

		_libdina_msqid=msgget(SHMSG_KEY, 0660 | IPC_CREAT);
		if(_libdina_msqid == -1)
			return -1;
	};

	/* Lock the shared memory */
	if(_libdina_lock() == -1)
		return -1;

	if(_libdina_rdonly) {
		sg = -1;
		errno = EEXIST;
	} else {
		/* Try to implicitly create the shared memory */
		sg = shmget(SHMEM_KEY, memsize,
			SHM_R | SHM_W	/* Read-write for us */
			| (SHM_R >> 3) | (SHM_W >> 3)
			| (SHM_R >> 6)	/* Read-only for the others */
			| IPC_CREAT | IPC_EXCL
		);
	};

	if(sg == -1) {
		if(errno != EEXIST) {
			_libdina_unlock();
			return -1;
		};

initlab:

		/* Try to get already allocated shared memory */
		sg = shmget(SHMEM_KEY, memsize, 0);

		if(sg == -1) {
			_libdina_unlock();
			return -1;
		};

		/* Attach memory to the user space */
		_libdina_mheader=(DINA_MemoryHeader *)
			shmat(sg, NULL, _libdina_rdonly?SHM_RDONLY:0);

		if(_libdina_mheader == (DINA_MemoryHeader *)-1
		|| !_libdina_mheader) {
			_libdina_unlock();
			return -1;
		};

		if(_libdina_mheader->version != DINA_IPC_VERSION) {
			_libdina_unlock();
			return -1;
		};

		if(_libdina_mheader->memsize != memsize) {
			/* Don't loop */
			if(mempasses++) {
				_libdina_unlock();
				return -1;
			};

			/* An unreal case (according to shmget(2)) */
			if(_libdina_mheader->memsize < memsize) {
				_libdina_mheader->memsize = memsize;
			} else {
				memsize = _libdina_mheader->memsize;
				shmdt(_libdina_mheader);
				_libdina_mheader = 0;
				goto initlab;
			};
		};
	} else {
		/* Attach memory to the user space */
		_libdina_mheader=(DINA_MemoryHeader *)shmat(sg, NULL, 0);
		if(_libdina_mheader == (DINA_MemoryHeader *)-1
		|| !_libdina_mheader) {
			_libdina_unlock();
			return -1;
		};

		/* Initialize memory */
		if(_libdina_init_memory() == -1) {
			_libdina_unlock();
			return -1;
		};
	};

	_libdina_lines = (LineEntry *)(_libdina_mheader
		+ sizeof(DINA_MemoryHeader));

	atexit(_libdina_cleanup);
	_libdina_initialized=1;

	_libdina_unlock();

	return 0;
};

/* Detach the shared memory segment */
void
_libdina_cleanup() {
	if(!_libdina_initialized)
		return;

	if(_libdina_mheader)
		shmdt(_libdina_mheader);

	_libdina_mheader = NULL;
	_libdina_lines = NULL;
	_libdina_sem = -1;
	_libdina_msqid = -1;
	_libdina_rdonly = 0;

	_libdina_initialized=0;
};

int _libdina_locks=0;

/* Lock the shared memory using the semaphore */
int
_libdina_lock() {
	int ret;
	struct sembuf op[2] = {
		{ 0, 0, 0 },
		{ 0, 1, SEM_UNDO }
	};

	if(_libdina_sem == -1)
		return -1;

	if(_libdina_locks)
		return 0;

	ret = semop(_libdina_sem, op, 2);

	if(ret == 0)
		_libdina_locks++;

	return ret;
};

/* Unlock the shared memory */
int
_libdina_unlock() {
	int ret;
	struct sembuf op[1] = { { 0,-1,0 } };

	if(_libdina_sem == -1)
		return -1;

	if(_libdina_locks == 0)
		return 0;

	ret = semop(_libdina_sem, op, 1);

	if(ret == 0)
		_libdina_locks--;

	return ret;
};


int
_libdina_init_memory() {
	DINA_MemoryHeader *mh = _libdina_mheader;

	if(!mh) return -1;

	memset(mh, 0, sizeof(DINA_MemoryHeader));

	time(&mh->created);
	mh->version = DINA_IPC_VERSION;
	mh->memsize = SHMEM_SIZE;

	return 0;
};


