#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/msg.h>
#include <sys/sem.h>

#include <libdina.h>
#include "intern.h"

void
_libdina_advertise_transition(LineEntry *le) {
	int n, v;
	int waterline = 0;
	u_short array[SHSEM_ITEMS];
	struct _msgp {
		long mtype;
		LineEntryEnvelope lenv;
	} msgp, tmp;
	struct msqid_ds dsbuf;
	register int copied=0;

	if(!le) return;

	if(semctl(_libdina_sem, SHSEM_ITEMS, GETALL, array) == -1)
		return;

	if(msgctl(_libdina_msqid, IPC_STAT, &dsbuf) == 0) {
		/* If dangerous situation, flag it. */
		if(dsbuf.msg_cbytes >= (dsbuf.msg_qbytes >> 1)) {
			waterline = 1;
			/* If filled more than 75%, here is red danger. */
			if((dsbuf.msg_qbytes - dsbuf.msg_cbytes)
			< (dsbuf.msg_qbytes >> 2))
				waterline = 2;
		};
	};

	for(n = 1; n < SHSEM_ITEMS; n++) {
		if(array[n] == 0) {
			/* Start reading unused messages by myself. */
			if(waterline > 0) {
				tmp.mtype = n;
				while(msgrcv(_libdina_msqid, &tmp,
					sizeof(LineEntryEnvelope), n, IPC_NOWAIT) == 0);
			};
			continue;
		} else {
			/* Start reading timed out messages by myself. */
			if(waterline > 1) {
				tmp.mtype = n;
				while(msgrcv(_libdina_msqid, &tmp,
					sizeof(LineEntryEnvelope), n, IPC_NOWAIT) == 0);
			};
		};

		msgp.mtype = n;
		if(!copied) {
			memcpy(&msgp.lenv.line, le, sizeof(LineEntry));
			msgp.lenv.Transition = _libdina_mheader->transitions;
			msgp.lenv.empty1=0;
			msgp.lenv.empty2=0;
			msgp.lenv.empty3=0;
			copied++;
		};

ret:
		v = msgsnd(_libdina_msqid, &msgp, sizeof(LineEntryEnvelope), 0);
		if(v == -1) {
			if(errno == EINTR)
				goto ret;
		};
	};
};

int
nas_line_set_connected(char *nas, int line, char *user, unsigned int sess) {
	LineEntry *iter;
	int n;

	if(_libdina_init())
		return -1;

	if(!nas || !user || line < 1)
		return -2;

	if(strlen(nas) > DI_NAS_ID
	|| strlen(user) > DI_USERNAME) {
		return -2;
	};

	/* Lock the shared memory */
	if(_libdina_lock() == -1)
		return -1;

	/*********** MEMORY LOCKED ****************/

	for(n = _libdina_mheader->lines,
		iter = _libdina_lines; n--; iter++) {
		/* Checking the line number first (speeds up searches) */
		if(iter->LineNumber != line) continue;
		/* Line found in those cases */
		if(!strncasecmp(iter->NAS, nas, DI_NAS_ID)) break;
		if(!*iter->NAS) break;
	};

	/* Line was not found */
	if(n < 0) {
		/* Add new line */
		iter = _libdina_lines + _libdina_mheader->lines;
		if(iter - _libdina_lines + sizeof(LineEntry) > SHMEM_SIZE) {
			errno = ENOMEM;
			_libdina_unlock();
			return -1;
		};
		iter->LineNumber = line;
		strncpy(iter->NAS, nas, DI_NAS_ID);
		_libdina_mheader->lines++;
	};

	time(&iter->TimeStamp);
	strncpy(iter->UserName, user, DI_USERNAME);
	iter->Transitions++;
	iter->SessionID=sess;

	/* Update line transitions counter */
	_libdina_mheader->transitions++;

	_libdina_advertise_transition(iter);

	/********** MEMORY UNLOCKED ****************/

	/* Unlock the shared memory */
	_libdina_unlock();
	return 0;
};

int
nas_line_set_disconnected(char *nas, int line, unsigned int sess) {
	LineEntry *iter;
	int n;

	if(_libdina_init())
		return -1;

	if(!nas || line < 1)
		return -2;

	if(strlen(nas) > DI_NAS_ID) {
		return -2;
	};

	/* Lock the shared memory */
	if(_libdina_lock() == -1)
		return -1;

	/*********** MEMORY LOCKED ****************/

	for(n = _libdina_mheader->lines,
		iter = _libdina_lines; n--; iter++) {
		/* Checking the session id or line number first */
		if(sess) {
			if(iter->SessionID != sess) continue;
		} else {
			if(iter->LineNumber != line) continue;
		};

		if(strncasecmp(iter->NAS, nas, DI_NAS_ID) != 0)
			continue;

		/* Line found! */
		*(iter->UserName) = '\0';
		time(&iter->TimeStamp);
		iter->Transitions++;
		break;
	};

	if( n < 0 ) {
		_libdina_unlock();
		return -1;
	};

	/* Update line transitions counter */
	_libdina_mheader->transitions++;

	_libdina_advertise_transition(iter);

	/********** MEMORY UNLOCKED ****************/

	/* Unlock the shared memory */
	_libdina_unlock();
	return 0;
};

