#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <sysexits.h>
#include <time.h>

#include <libdina.h>

void usage();

char *user=NULL;
char *nas=NULL;
int line=-1;

int rdonly = 1;
int watch_trans = 0;
int quiet=0;

int Count(int);
int Iterate(int, LineEntry *);
int showall = 0;
void watch_transitions();

int
main(int ac, char **av) {
	int ch;
	int n;
	time_t tloc;

	while((ch = getopt(ac, av, "Iwaqu:n:l:")) != -1)
	switch(ch) {
		case 'I':
			if(getuid() && getuid() && getgid() && getegid())
				errx(1, "Super-user or group privileges required.");
			rdonly = 0;
			break;
		case 'q':
			quiet++;
			break;
		case 'a':
			showall=1;
			break;
		case 'u':
			if(user)
				warnx("Username specified more than once.");
			user=optarg;
			break;
		case 'n':
			if(nas)
				warnx("NAS ID specified more than once.");
			nas=optarg;
			break;
		case 'l':
			if(line >= 0)
				warnx("Line specified more than once.");
			if(*optarg < '0' || *optarg > '9')
				errx(1, "Line value must be numeric");
			line=atoi(optarg);
			break;
		case 'w':
			watch_trans = 1;
			break;
		default:
			usage();
	};

	ac -= optind;
	av += optind;

	if(rdonly && dina_init_readonly()) {
		warnx("Can't init DINA in read-only mode.");
		return -1;
	};

	tloc = (time_t)dina_ctl_get(DCTL_IPC_CREATED);
	n = (int)dina_ctl_get(DCTL_LINES_COUNT);
	if(tloc < 0 || n == -1)
		err(1, "Could not initialize DINA engine.");

	if(!quiet) {
		fprintf(stderr, "Since: %s", ctime(&tloc));
		fprintf(stderr, "Lines defined: %d\n", n);
	};

	if(watch_trans)
		watch_transitions();

	n = iterate_through_lines(user, nas, line, quiet?NULL:Count, Iterate);
	if(n == -1)
		errx(1, "Can't get lines information.");

	if(!quiet && n > 0)
	fprintf(stderr, "Lines filtered: %d\n", n);

	return 0;
};

int Count(int n) {
	if(n <= 0 || quiet != 0)
		return n;

	fprintf(stderr,
		"No. S %-16s %-16s %s %s %s\n",
		"Username",
		"NAS",
		"Line",
		"Sess",
		"Transition date"
	);

	return n;
};

int
Iterate(int n, LineEntry *le) {
	int q;
	q = *le->UserName;
	if(!q) {
		if(!showall)
			return 0;
		*le->UserName = '?';
	};
	printf("%2d. %s %-16s %-16s %-4d %-4d %s",
		n,
		q?"*":" ",
		le->UserName,
		le->NAS,
		le->LineNumber,
		le->SessionID,
		ctime(&le->TimeStamp)
	);

	return 0;
};

void
usage() {
	fprintf(stderr, "xw (c) 1999, 2000 Lev Walkin <vlm@spelio.net.ru>\n"
		"Usage: xw [-I] [-q] [-a] [-w] [-u <username>] [-n <nas>] [-l <line_num>]"
	);
	exit(EX_USAGE);
};

void
watch_transitions() {
	int status;
	char buf[sizeof(LineEntry)];
	LineEntryEnvelope le;
	int transitions=0;
	int ignorefirst=1;

	if(dina_next_transition(NULL, 0))
		errx(2, "Can't register transitions handler.");

	while(1) {
		if(!quiet)
			fprintf(stderr, "[Waiting for line transition...]\n");
		if(dina_next_transition(&le, 0))
			errx(1, "Error getting next line transition.");

        if(le.Transition - transitions > 1) {
			if(ignorefirst)
				ignorefirst=0;
			else
				fprintf(stderr, "[WARNING: %d previous transitions missed!]\n",
					le.Transition-transitions);
		};
        transitions=le.Transition;

		printf("Date:\t%s", ctime((time_t *)&(le.line.TimeStamp)));

		if(!*le.line.UserName) { *le.line.UserName = '?'; status = 0; }
		else { status = 1; };
		printf("Status:\t%s\n", status?"Connected":"Disconnected");

		snprintf(buf, DI_NAS_ID, "%s", le.line.NAS);
		printf("Nas:\t%s\n", buf);
		printf("Line:\t%d (transitions: %u)\n", le.line.LineNumber, le.line.Transitions);
		snprintf(buf, DI_USERNAME, "%s", le.line.UserName);
		printf("User:\t%s\n", buf);
		printf("\n");
	};

	exit(0);
};

